/*
    Meter
 
    Electriciteits meter programma voor de Arduino Ethernet

    Copyright (c) Martijn van der Horst 2012

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include <SD.h>
#include <Time.h>
#include <SPI.h>
#include <Ethernet.h>

/* Instellingen:*/
const boolean use_netwerk = true;            // Gebruik de netwerk verbinding (true = ja, false = nee)
const boolean use_dhcp = true;               // Automatisch de netwerk instellingen ophalen? (true = ja, false = nee)
const boolean use_sd = true;                 // Gebruik maken van de SD kaart? (true = ja, false = nee)
const unsigned int tijd_ophalen = 3600;      // Om de hoeveel seconden moet de tijd opnieuw opgehaald worden van internet?
const int led = 9;                           // De pin met de on-board LED er aan (pin 9 op de Arduino Ethernet)
const double kWhperOmwenteling = 1.0/187.5;  // Hoeveel kWh is er gebruikt elke keer dat de rode streep voorbij komt?
const unsigned long W1ms = kWhperOmwenteling * 3600000000.0; // Hoevel W is er gebruikt als een omwenteling 1 ms zou duren?
const int chipSelect = 4;                    // De pin voor de SD kaart (pin 4 op de Arduino Ethernet)
const int sensor = 8;                        // De pin met de sensor er aan (Pas op: het afstellen van de interrupt routines in "setup" neemt aan dat dit 8 is)
const int bestandsnaamLengte = 13;           // Maximum lengte van een bestandsnaam (incl '\0')
const int tijdLengte = 20;                   // Maximum lengte van een tijdstip omgezet naar string (incl '\0')
const unsigned int lokaleNTPPoort = 8888;    // Lokale poort om te luisteren naar UDP pakketten van de NTP server (wordt gebruikt om de tijd van internet op te halen)
const int HTTPregelbuffer=20;                // Maximale lengte van een regel in een HTTP request (de rest van een langere regel wordt genegeerd)
const int seconden_per_minuut = 60;          // Het aantal seconden in een minuut
const int seconden_per_uur = 60*60;          // Het aantal seconden in een uur
const int seconden_per_dag = 24*60*60;       // Het aantal seconden in een dag
const int NLzomertijd = 2*60*60;             // De nederlandse zomertijd loopt 2 uur voor op UTC
const int NLwintertijd = 1*60*60;            // De nederlandse wintertijd loopt 1 uur voor op UTC

// Het IP adres van de NTP server om de tijd van op te halen
IPAddress NTPServer(213, 125, 225, 50); // nl.pool.ntp.org NTP server (op 19-11-2012)

// MAC adres van de arduino:
byte mac[] = {  
  0x90, 0xA2, 0xDA, 0x0D, 0x8D, 0xEF };

// Het IP adres van de arduino (als de instellingen niet automatisch worden opgehaald)
IPAddress ip(192,168,0,11);

/* Variabelen:*/
volatile int sensorRauw = LOW;             // De huidige, ongefilterde, waarde van de sensor
volatile int sensorWaarde = LOW;           // De huidige, gefilterde, waarde van de sensor (het resultaat van de ruis filter)

unsigned long laatsteOmwenteling = 0;      // Tijdstip van de laatste omwenteling (wordt gebruikt om het verbruik te berekenen)
unsigned long huidigeOmwenteling = 0;      // Tijdstip van de huidige omwenteling (wordt gebruikt om het verbruik te berekenen)
volatile unsigned int verbruik = 0;        // Het huidige verbruik in Watt (wordt bijgewerkt na iedere omwenteling)
volatile unsigned long omwentelingen = 0;  // Het aantal omwentelingen sinds het programma begonnen is
volatile boolean nieuwe_gegevens = false;  // Of er een nieuwe omwenteling is geweest sinds deze variabele op "FALSE" werdt gezet

unsigned long energieMinuut = 0;           // De hoeveelheid gebruikte Joule in de huidige minuut
time_t energieBijgewerkt = 0;              // Tijdstip tot waarop "energieMinuut" is bijgewerkt

time_t laatsteTijdSynchronisatie = 0;      // Laatste keer dat de tijd van internet is opgehaald

EthernetServer server(80);                 // Web server
EthernetUDP Udp;                           // Een UDP instatie voor het ophalen van de huidige tijd van internet

char tijd[tijdLengte];                     // Deze variabele bevat de laatste tijd die naar een string geconverteerd is
char bestandsnaam[bestandsnaamLengte];     // Deze variabele bevat de laatste gegenereerde bestandsnaam

// De setup routine wordt 1 keer uitgevoerd na een reset:
void setup()  { 
  // Open de seriele poort voor communicatie:
  Serial.begin(9600);
  Serial.println("Start..."); 

  if (use_netwerk) {
    // Start de netwerk verbinding:
    Serial.println("Netwerk... ");
    if (use_dhcp) {
      Ethernet.begin(mac);
    } 
    else {
      Ethernet.begin(mac, ip);
    }
    server.begin();
    Serial.print("IP adres ");
    Serial.println(Ethernet.localIP());

    // Haal de tijd op
    Udp.begin(lokaleNTPPoort);
    tijdBijwerken(true);
  }

  // Maak contact met de SD kaart
  if (use_sd) {
    Serial.print("SD kaart... ");
    pinMode(10, OUTPUT); // Pin 10 moet uivoer zijn van de SD bibliotheek
    if (!SD.begin(chipSelect)) {
      Serial.println("FOUT.");
    } 
    else {
      Serial.println("OK.");
    }
  }

  // Algemene instellingen
  Serial.print("Alg. instellingen... ");

  // Pin 3 wordt een uitvoer met een 36kHz signaal er op
  pinMode(3, OUTPUT);
  TCCR2A = bit(COM2B1) | bit(WGM21) | bit(WGM20);
  TCCR2B = bit(WGM22) | bit(CS21);
  OCR2A = 54; /* Frequentie = 16Mhz/8/(OCR2A+1)=36kHz */
  OCR2B = 27; /* (OCR2B+1)/(OCR2A+1)=50% duty cycle */

  // Stel de pinnen voor de led en de sensor in
  pinMode(led, OUTPUT);
  pinMode(sensor, INPUT);

  // Tijds afhankelijke instellingen
  energieBijgewerkt = now();

  // Stel de interrupt routines in
  PCICR |= bit(PCIE0);    // Interrupt enable voor pin change interrupt op pin 8 t/m 13
  PCMSK0 |= bit(PCINT0);  // Luister alleen naar pin 8 voor de interrupt
  TCCR1A = 0;             // Timer 1 in normale modus
  TCCR1B = bit(CS11);     // Timer 1 gaat op 16Mhz/8=2Mhz lopen, Timer 1 is 16-bits, een overflow zal dus iedere (2^16)/2Mhz=32.8ms plaatsvinden (blz 137 van datasheet)
  interrupts();           // Zet interrupts aan

  Serial.println("OK.");
} 

// De interrupt routine wordt uitgevoerd als de sensor iets registreerd, d.w.z. van waarde veranderd
ISR(PCINT0_vect){
  // Lees de sensor:
  int sensorState = digitalRead(sensor);
  sensorRauw = sensorState;

  // Meteen het ledje aansturen met de meting, zonder ruis filtering
  digitalWrite(led, sensorState);

  if (sensorState != sensorWaarde) {
    // De sensor is van waarde veranderd
    // als deze waarde stabiel is, dan moeten we daar iets mee doen, dus we plannen een interrupt
    TCNT1 = 0;                      // Zet de timer op 0
    TIFR1 |= bit(TOV1);             // Reset de overflow interrupt vlag
    TIMSK1 |= bit(TOIE1);           // Zet de overflow interrupt aan
  } 
  else {
    // De sensor heeft weer zijn oude waarde
    // De waarde was dus niet stabiel, dus we stoppen de geplande interrupt
    TIMSK1 &= ~bit(TOIE1);          // Zet de overflow interrupt uit
    TIFR1 |= bit(TOV1);             // Reset de overflow interrupt vlag
  }
}

// De interrupt routine die wordt uitgevoerd als de sensor al een tijdje stabiel is en anders is dan de huidige sensorWaarde (m.b.v. de PCINT0_vect interrupt routine)
ISR(TIMER1_OVF_vect){
  // De huidige, ongefilterde, status van de sensor is stabiel, we kunnen 'm gebruiken als nieuwe sensor waarde:
  int sensorState = sensorRauw;
  sensorWaarde=sensorState;

  // De interrupt is afgehandeld, zet de volgende interrupt uit:
  TIMSK1 &= ~bit(TOIE1);          // Zet de overflow interrupt uit
  TIFR1 |= bit(TOV1);             // Reset de overflow interrupt vlag

  if (sensorState == HIGH) {
    // We zijn van laag naar hoog gegaan, de rode streep zit nu dus net niet meer voor de sensor
    laatsteOmwenteling = huidigeOmwenteling; // Onthoud het tijdstip dat de rode streep voor de sensor kwam
  } 
  else { // (sensorState == LOW)
    // We zijn van hoog naar laag gegaan, de rode streep zit nu dus voor de sensor
    huidigeOmwenteling = millis(); // Onthoud het tijdstip van deze omwenteling, d.w.z. ht tijdstip van wanneer de rode streep voor de sensor is gekomen

    // Nieuwe omwenteling
    omwentelingen++;

    // Bereken het verbruik
    unsigned long tijd = (huidigeOmwenteling - laatsteOmwenteling); // Tijd in milliseconden tussen deze en de vorige omwenteling
    verbruik = W1ms/tijd; // Verbruik in W
    nieuwe_gegevens = true; // Er zijn nieuwe gegevens; het verbruik is bijgewerkt
    
    // Print het verbruik naar de seriele poort: (Waarschuwing: niet te veel printen in een interrupt routine!)
    /*Serial.print(verbruik);
    Serial.println(" W");
    Serial.print(tijd);
    Serial.println(" ms");*/

  }

}

// Een functie die een nummer naar een tekst van 2 cijfers omzet
void nr_naar_2cijfers(const int nr, char buffer[2])
{
  buffer[0]='0'+((nr/10)%10);
  buffer[1]='0'+(nr%10);
}

// Een functie die een nummer naar een tekst van 4 cijfers omzet
void nr_naar_4cijfers(const int nr, char buffer[4])
{
  nr_naar_2cijfers(nr/100,buffer);
  nr_naar_2cijfers(nr%100,buffer+2);
}

// Een functie die een tijdstip (datum, uur en minuut) omzet naar een string en in de globale variabele "tijd" zet
void tijd_als_string(const time_t t)
{
  // Basis layout:
  strcpy(tijd,"yyyy-mm-dd hh:mm");

  // Invullen:
  nr_naar_4cijfers(year(t), tijd+0);   // Het jaartal
  nr_naar_2cijfers(month(t), tijd+5);  // De maand
  nr_naar_2cijfers(day(t), tijd+8);    // De dag
  nr_naar_2cijfers(hour(t), tijd+11);  // Het uur
  nr_naar_2cijfers(minute(t), tijd+14);// De minuut
}

// Een functie die een tijdstip (datum, uur en minuut) omzet naar een string en in de globale variabele "tijd" zet
void tijd_als_string_voor_javascript(const time_t t)
{
  // Basis layout:
  strcpy(tijd,"(yyyy,mm,dd,hh,mm)");

  // Invullen:
  nr_naar_4cijfers(year(t), tijd+1);   // Het jaartal
  nr_naar_2cijfers(month(t)-1, tijd+6);// De maand-1 (want Javascript begint bij 0 te tellen, d.w.z. Januari is maand 0)
  nr_naar_2cijfers(day(t), tijd+9);    // De dag
  nr_naar_2cijfers(hour(t), tijd+12);  // Het uur
  nr_naar_2cijfers(minute(t), tijd+15);// De minuut
}

// Een functie die een bestandsnaam voor het maandelijkse CSV bestand  geeft gebaseerd op een tijdstip
void naam_csv_bestand(const time_t t)
{
  // Basis layout:
  strcpy(bestandsnaam, "yyyy-mm.csv");

  // Invullen:
  nr_naar_4cijfers(year(t), bestandsnaam+0);   // Het jaartal
  nr_naar_2cijfers(month(t), bestandsnaam+5);  // De maand
}

// Een functie die een bestandsnaam voor het dagelijkse webserver bestand  geeft gebaseerd op een tijdstip
void naam_dag_bestand(const time_t t)
{
  // Basis layout:
  strcpy(bestandsnaam, "yyyy-mm-dd.dag");

  // Invullen:
  nr_naar_4cijfers(year(t), bestandsnaam+0);   // Het jaartal
  nr_naar_2cijfers(month(t), bestandsnaam+5);  // De maand
  nr_naar_2cijfers(day(t), bestandsnaam+8);    // De dag
}

// Geeft aan dat een bestand niet geopend kon worden
void fout_bij_openen_bestand()
{
  Serial.print("FOUT bij openen \"");
  Serial.print(bestandsnaam);
  Serial.println("\"");
}

// Functie om de bestanden op de SD kaart bij te werken
void SDBijwerken() {
  // Kijk of er nieuwe gegevens zijn binnen gekomen van de interupt routine
  if (nieuwe_gegevens) {
    time_t nu = now(); // De huidige tijd
    int bijtewerken = nu - energieBijgewerkt; // Hoe veel seconden loopt de administratie achter?

    Serial.print(verbruik);
    Serial.println(" W");

    // Zijn we in een nieuwe dag beland?
    /*if (day(nu) != day(energieBijgewerkt)) {  
      // We gaan beginnen aan een nieuwe dag

      // Zorg dat het bestand voor de nieuwe dag wordt weggehaald,
      // als het bestaat dan bevat dit nl. nog de gegevens van vorige maand
      naam_dag_bestand(nu);
      if (use_sd && SD.exists(bestandsnaam)) SD.remove(bestandsnaam);
    }*/
  
    // Werk de administratie bij indien nodig
    while (bijtewerken > 0) {
      // De administratie loopt minstens 1 seconde achter, dus er moet bijgewerkt worden
  
      if (minute(nu) == minute(energieBijgewerkt)) {
  
        // We zitten wel al in de goede minuut, maar lopen wel 1 of meerdere seconden achter. Werk de administratie bij:
        energieMinuut = energieMinuut + (bijtewerken * verbruik);
  
        // We zijn helemaal klaar met het bijwerken van de administratie:
        bijtewerken = 0;
  
      } 
      else {
  
        // We lopen 1 of meerdere minuten achter
        int secondenOver = 60 - second(energieBijgewerkt); // Het aantal seconden dat nog over was in de minuut waar we mee bezig waren
  
        // Werk de administratie van de oorspronkelijke minuut bij
        energieMinuut = energieMinuut + (secondenOver * verbruik);
  
        // Schrijf het het resultaat van de minuut naar de SD card
        if (use_sd) {

          // Werk het maandelijkse CSV bestand bij
          naam_csv_bestand(energieBijgewerkt);
          File dataFile = SD.open(bestandsnaam, FILE_WRITE); // Open het bestand om te schrijven
          if (dataFile) {
            // Het bestand is geopend, schrijf de data:
            tijd_als_string(energieBijgewerkt);
            dataFile.print(tijd);  // Schrijf de minuut naar het bestand
            dataFile.print(',');
            dataFile.println(energieMinuut/60);      // Schrijf het gemiddelde verbruik over de minuut in Watt naar het bestand
            dataFile.close();
          } 
          else fout_bij_openen_bestand();
          
          // Werk het dagelijkse webserver bestand bij
          naam_dag_bestand(energieBijgewerkt);
          dataFile = SD.open(bestandsnaam, FILE_WRITE); // Open het bestand om te schrijven
          if (dataFile) {
            // Het bestand is geopend, schrijf de data:
            tijd_als_string_voor_javascript(energieBijgewerkt);
            dataFile.print("[new Date");
            dataFile.print(tijd);
            dataFile.print(',');
            dataFile.print(energieMinuut/60);        // Schrijf het gemiddelde verbruik over de minuut in Watt naar het bestand
            dataFile.println("],");
            dataFile.close();
          } 
          else fout_bij_openen_bestand();

        }
  
        // De administratie is bijgewerkt met "secondenOver" seconden, en we beginnen aan een nieuwe minuut
        bijtewerken = bijtewerken - secondenOver;
        energieBijgewerkt = energieBijgewerkt + secondenOver;
        energieMinuut = 0;
      }
    }

    // Alle gegevens zijn verwerkt
    energieBijgewerkt = nu;
    // We gaan wachten tot de interrupt routine nieuwe gegevens levert
    nieuwe_gegevens = false;
  }
}

// Geeft het begin van de zomertijd in nederland terug voor een bepaald jaar
time_t beginZomertijdinNL(int jaar) {
  tmElements_t tm;
  time_t t;

  // De zomertijd gaat in op de laatste zondag van maart om 2 uur 's nachts lokale tijd, dat is 1 uur 's nachts in UTC
  // Dus zoeken we eerst naar de eerste zondag van april:
  tm.Hour = 1;            // Een uur 's nachts
  tm.Minute = 0;
  tm.Second = 0;
  tm.Day = 1;             // Eerste dag van
  tm.Month = 4;           // April
  tm.Year = jaar - 1970;  // In het huidige jaar
  t = makeTime(tm);       // Zet tm om naar een tijd in seconden
  // Zoek nu de eerste zondag van de maand:
  t = t + ((1 - weekday(t) + 7) % 7) * seconden_per_dag;
  // En gaan vanaf die eerste zondag 7 dagen terug, dan zitten we op de laatste zondag in maart
  t = t - 7 * seconden_per_dag;

  return t;
}

// Geeft het einde van de zomertijd in nederland terug voor een bepaald jaar
time_t eindeZomertijdinNL(int jaar) {
  tmElements_t tm;
  time_t t;

  // De zomertijd stopt op de laatste zondag van oktober om 3 uur 's nachts lokale tijd, dat is 1 uur 's nachts in UTC
  // Dus zoeken we de eerste zondag van november:
  tm.Hour = 1;            // Een uur 's nachts
  tm.Minute = 0;
  tm.Second = 0;
  tm.Day = 1;             // Eerste dag van
  tm.Month = 11;          // November
  tm.Year = jaar - 1970;  // In het huidige jaar
  t = makeTime(tm);       // Zet tm om naar een tijd in seconden
  // Zoek nu de eerste zondag van de maand:
  t = t + ((1 - weekday(t) + 7) % 7) * seconden_per_dag;
  // En gaan vanaf die eerste zondag van 7 dagen terug, dan zitten we op de laatste zondag in oktober
  t = t - 7 * seconden_per_dag;

  return t;
}

// Functie die aangeeft of deze gegeven tijd (in UTC) in de nederlandse zomertijd valt.
boolean isZomerTijd(time_t utc_tijd) {
  // Om dit te bepalen zoeken we naar het eerstvolgende tijdstip waarop de klok verzet zal worden

  // De zomertijd gaat in op de laatste zondag van maart om 2 uur 's nachts lokale tijd, dat is 1 uur 's nachts in UTC
  // De zomertijd stopt op de laatste zondag van oktober om 3 uur 's nachts lokale tijd, dat is 1 uur 's nachts in UTC

  // Dus zoeken we eerst naar de eerst volgende keer dat het 1 uur 's nachts op zondag zal worden:
  utc_tijd += ((60-second(utc_tijd)) % 60);                     // We zitten nu op de eerstvolgende hele minuut
  utc_tijd += ((60-minute(utc_tijd)) % 60)*seconden_per_minuut; // We zitten nu op het eerstvolgende hele uur
  utc_tijd += ((25-hour(utc_tijd))%24)*seconden_per_uur;        // We zitten nu op het eerstvolgende 1 uur 's nachts
  utc_tijd += ((8 - weekday(utc_tijd)) % 7) * seconden_per_dag; // We zitten nu op het eerstvolgende zondag om 1 uur 's nachts

  
  // Als de eerstvolgende zondag om 1 uur 's nachts valt in de maanden januari, februari of maart,
  // dan zitten we dus nog voor  de laatste zondag in maart en dus in de wintertijd
  // Als de eerstvolgende zondag om 1 uur 's nachts valt in de maanden april, mei, juni, juli, augustus, september of oktober,
  // dan zitten we dus nog voor de laatste zondag in oktober en dus in de zomertijd
  // Als de eerstvolgende zondag om 1 uur 's nachts valt in de maanden november of december,
  // dan zitten we dus na de laatste zondag in kotober en dus weer in de wintertijd

  // Kijk in welke maand deze zondag valt, en of het zomertijd is:
  int maand = month(utc_tijd);
  return ((maand >= 4) && (maand <= 10));
}

// Functie voor het omrekenen van UTC tijd naar lokale (NL) tijd
time_t lokaleTijduitUTC(time_t utc_tijd) {
  time_t lokaleTijd;

  // Bereken de tijden waarop de zomertijd begint en eindigt in UTC:
  /*time_t startZomertijd = beginZomertijdinNL(year(utc_tijd));
  time_t stopZomertijd = eindeZomertijdinNL(year(utc_tijd));

  // Kijk of we in zomertijd zitten:
  boolean zomertijd = (utc_tijd >= startZomertijd && utc_tijd < stopZomertijd);*/
  
  // Reken de lokale tijd uit
  if (isZomerTijd(utc_tijd))
      lokaleTijd = utc_tijd + NLzomertijd;  // Nederlandse zomertijd is UTC + 2 uur
  else
      lokaleTijd = utc_tijd + NLwintertijd; // Nederlandse wintertijd is UTC + 1 uur
  
  return lokaleTijd;
}

// Functie om de tijd op te halen van internet
// De forceer vlag geeft aan of de tijd op gehaald moet worden (true), of dat het alleen hoeft als er al een tijdje niets is gedaan (false)
void tijdBijwerken(boolean forceer) {
  // Kijk of we om een nieuwe tijd van internet moeten vragen:
  if ((forceer) || ((now() - laatsteTijdSynchronisatie) > tijd_ophalen)) {
    const int NTP_PACKET_SIZE= 48; // De tijd zit in de eerste 48 bytes van een NTP bericht
    byte packetBuffer[ NTP_PACKET_SIZE] = {  0  }; //Buffer voor de inkomende en uitgaande berichten

    // We gaan een verzoek versturen naar de NTP server
    // Noot: packetBuffer staat al vol met nullen, dus we zetten verder alleen wat nodig is:
    packetBuffer[0] = 0b11100011;  // LI (klok nog niet gesynchroniseerd), Versie (4), Mode (client)
    packetBuffer[1] = 0;           // Stratum ofwel klok type (ongedefinieerd/niet beschikbaar)
    packetBuffer[2] = 6;           // Polling Interval (2^6 seconden)
    packetBuffer[3] = 0xEC;        // Precisie van de klok (-20 -> 2^(-20) seconden, ofwel ca. 1 microseconde)
    // Root Delay & Root Dispersion (8 bytes in totaal) staan op 0
    packetBuffer[12]  = 49;        // '1'
    packetBuffer[13]  = 0x4E;      // 'N'
    packetBuffer[14]  = 49;        // '1'
    packetBuffer[15]  = 52;        // '4'

    // Het NTP pakket is compleet
    Udp.beginPacket(NTPServer, 123);          // NTP verzoeken moeten naar poort 123
    Udp.write(packetBuffer,NTP_PACKET_SIZE);  // Verstuur het pakket
    Udp.endPacket();                          // Geef aan dat dat het hele pakket was

    // Wacht een seconde om de server tijd te geven om te antwoorden
    delay(1000);

    // Kijk of er een antwoord van de NTP server is binnen gekomen
    if ( Udp.parsePacket() ) {

      // Er is een pakket binnen gekomen, we lezen de gegevens in:
      Udp.read(packetBuffer,NTP_PACKET_SIZE);  // Plaats de data in de pakket buffer
  
      // De tijd staat in het ontvangen pakketje, de tijd begint op byte 40 en is 4 bytes groot
      unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
      unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);  
      // Combineer de vier bytes tot een enkel getal, dit is de tijd volgens NTP (UTC tijd in seconden sinds 1 januari 1900):
      unsigned long secsSince1900 = highWord << 16 | lowWord;  
  
      // Converteer de NTP tijd naar tijd die gebruikt wordt door de Time bibliotheek (Unix tijd):
      // Unix tijd begint op 1 januari 1980. Het verschil in seconden is 2208988800:
      const unsigned long seventyYears = 2208988800UL;     
      // Het verschil:
      unsigned long epoch = secsSince1900 - seventyYears;  
  
      // Print de tijd
      Serial.print("UTC : ");
      tijd_als_string(epoch);
      Serial.println(tijd);
      
      // Converteer naar lokale tijd
      time_t lokaleTijd = lokaleTijduitUTC(epoch);
      
      // Sla de nieuwe tijd op
      setTime(lokaleTijd);
      laatsteTijdSynchronisatie = lokaleTijd;

      // Zorg dat de energie administratie niet in de war raakt (de energie administratie kan niet voor lopen):
      if (energieBijgewerkt > lokaleTijd) energieBijgewerkt = lokaleTijd;
      
      // Print de locale tijd
      Serial.print("NL tijd: ");
      tijd_als_string(now());
      Serial.println(tijd);   
      
    }
  }
}

// Stuur het begin van een antwoord naar de web browser
void stuurHeader(EthernetClient *client, const char *response, const char* content_type) {
  client->print("HTTP/1.1 ");
  client->println(response);
  if (content_type != 0) {
    client->print("Content-Type: ");
    client->println(content_type);
  }
  client->println("Connection: close");
  client->println();
}

// Stuur een bestand van de SD kaart naar de web browser
void stuurBestand(EthernetClient *client) {
  if (use_sd) {
    File dataFile = SD.open(bestandsnaam, FILE_READ); // Open het bestand om te lezen
    const size_t buffer_grootte = 64;
    uint8_t buffer[2*buffer_grootte+10];
    size_t buffer_index = 0;

    if (dataFile) {
      // Het bestand is geopend, stuur het door naar de web browser
      int c;
            
      while ((c = dataFile.read()) >= 0) {
        // Er zijn nog bytes te versturen
        client->print((char)c);
        /*buffer[buffer_index]=(uint8_t)c;
        buffer_index++;
        if (buffer_index >= buffer_grootte) {
          buffer_index = 0;
          client->write(buffer, buffer_grootte);
        }*/
      }

      //if (buffer_index > 0) client->write(buffer, buffer_index);

      dataFile.close();  
    }
  }
}


// Stuur de hoofd index van de website  naar de web browser
void stuurIndexBestand(EthernetClient *client) {
  if (use_sd) {
    File dataFile = SD.open("index.htm", FILE_READ); // Open het bestand om te lezen

    if (dataFile) {
      // Het bestand is geopend, stuur het door naar de web browser
      int c;

      // Vertel de web browser eerst wat voor bestand we gaan sturen
      stuurHeader(client, "200 OK", "text/html");

      while ((c = dataFile.read()) >= 0) {
        // De variable c bevat de volgende byte van het bestand, en
        // we blijven doorgaan zolang we nog niet het hele bestand gehad hebben

        if (c == '#') {
          // Als het bestand een # teken bevat dan gaan we dat stukje vervangen door iets anders
          // wat dat anders is, hangt af van het getal dat achter het #-teken staat:
          int vervang = dataFile.parseInt();
          switch(vervang) {
            
            // Tijd gerelateerd
            case 10: // Huidige tijd
              tijd_als_string(now());
              client->print(tijd);
              break;
            case 11: // Laatste tijd synchronisatie
              tijd_als_string(laatsteTijdSynchronisatie);
              client->print(tijd);
              break;
              
            // Energie gerelateerd
            case 20: // Huidig verbruik
              client->print(verbruik);
              break;
            case 21: // Aantal omwentelingen
              client->print(omwentelingen);
              break;
            case 22: // Laatste omwenteling
              tijd_als_string(energieBijgewerkt);
              client->print(tijd);
              break;

            // Grote hoeveelheden data:
            case 100: // Het bestand met het verbruik van vandaag
              naam_dag_bestand(now());
              stuurBestand(client);
              break;
            case 101: // Het bestand met het verbruik van gisteren
              naam_dag_bestand(now()-seconden_per_dag);
              stuurBestand(client);
              break;
            default:
              // Onbekend nummer, doe niets.
              client->print((char)c);
          }
        } else {
          // Deze byte gaan we gewoon doorsturen:
          client->print((char)c);
          //Serial.print((char)c);
        }
      }
    }
    dataFile.close();  
  }
}

// Handel vragen aan de web server af
void webServer() {
  if (use_netwerk) {
    // Kijk of het netwerk adres bijgewerkt moet worden
    // FIXME: Ethernet.maintain();
    
    // Kijk of er verbindingen zijn
    EthernetClient client = server.available();
    if (client) {
      Serial.println("browser");
      // Lees het HTTP request in
      char regel[HTTPregelbuffer+1];
      int index = 0;
      boolean bestandsnaamGevonden = false;
      regel[0] = 0;

      while (client.connected()) {
        if (client.available()) {
          // Lees 1 byte
          char c = client.read();
          // Print de nieuwe byte
          Serial.write(c);

          // Bekijk de byte
          if (c == '\n' && (index == 0)) {
            // Het einde van de regel is bereikt, en de regel is leeg
            // dit betekend dat we aan het einde van het HTTP request zijn
            // nu moet het verwerkt worden:

            if (bestandsnaamGevonden) {
              // De web-browser heeft om een specified bestand gevraagd
              if (use_sd && SD.exists(bestandsnaam)) {
                // Het bestand bestaat, dus stuur het op
              
                // Vertel de web browser eerst dat we een bestand gaan sturen
                stuurHeader(&client, "200 OK", 0);
              
                // Stuur dan het bestand
                stuurBestand(&client);
              } else {
                // Het bestand bestaat niet, stuur een foutmelding op
                stuurHeader(&client, "404 Not Found", 0);
              }
            } else {
              // Stuur het index.htm bestand
              stuurIndexBestand(&client);
              /*stuurHeader(&client, "200 OK", 0);
              tijd_als_string(now());
              client.println("<html>");
              client.println(tijd);
              client.println("</html>");*/
            }
            
            // Spring uit de while-lus
            break;

          }
          if (c == '\n') {
            // We beginnen met een nieuwe regel
            
            // Kijk of de huidige regel een vraag naar een bestandsnaam bevat
            if ((strstr(regel, "GET / ") == 0) && (strstr(regel, "GET /") != 0)) {
              // De regel bevat een "GET" request, en het is niet een request voor de hoofd-pagina, maar voor een specifiek bestand
              char *einde = strstr(regel+5," "); // Het einde van de bestandsnaam (de eerste spatie na "GET /")
              if (einde != 0) *einde=0;       // Kap de regel af net na het einde van de bestandsnaam

              // Kopieer de bestandsnaam uit de regel, naar de variabele "bestandsnaam"
              strncpy(bestandsnaam, regel+5, bestandsnaamLengte);
              
              // Geef aan dat we een bestandsnaam gevonden hebben
              bestandsnaamGevonden=true;
            }
            
            // Zet de index aan het begin van de buffer, om de nieuwe regel op te gaan slaan
            index = 0;
            regel[0]=0;
          } 
          else if (c != '\r') {
            // De byte bevat een tekst karakter, voeg het toe aan de huidige regel:
            regel[index] = c;
            // De volgende byte zal op de volgende plaats in de buffer gaan
            index=min(index+1, HTTPregelbuffer);
            // De buffer moet altijd eindigen met een 0
            regel[index] = 0;
          }
        }       
      }
      // Geef de web browser tijd om de data te ontvangen
      delay(1);
      // Gooi overgebleven data weg
      client.flush();
      delay(1);
      // Sluit de verbinding
      client.stop();
      Serial.println("browser gestopt");
    }
  }
}

// De loop routine wordt elke keer weer opnieuw uitgevoerd:
void loop()  {
  SDBijwerken();

  tijdBijwerken(false);

  webServer();
}


